package org.nuxeo.ecm.core.jcr.search.jrplug;

import java.io.Reader;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.RepositoryException;

import org.apache.jackrabbit.core.PropertyId;
import org.apache.jackrabbit.core.query.TextFilter;
import org.apache.jackrabbit.core.query.lucene.FieldNames;
import org.apache.jackrabbit.core.query.lucene.NamespaceMappings;
import org.apache.jackrabbit.core.state.ItemStateException;
import org.apache.jackrabbit.core.state.ItemStateManager;
import org.apache.jackrabbit.core.state.NoSuchItemStateException;
import org.apache.jackrabbit.core.state.NodeState;
import org.apache.jackrabbit.core.state.PropertyState;
import org.apache.jackrabbit.core.value.InternalValue;
import org.apache.jackrabbit.name.NoPrefixDeclaredException;
import org.apache.jackrabbit.name.QName;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class AdvNodeIndexer extends LuceneUnalterredNodeIndexer {

    /**
     * The logger instance for this class.
     */
    private static final Logger log = LoggerFactory.getLogger(AdvNodeIndexer.class);

    /**
     * Creates a Lucene Document from a node.
     *
     * @param node          the node state to index.
     * @param stateProvider the state provider to retrieve property values.
     * @param mappings      internal namespace mappings.
     * @param textFilters   list of text filters to use for indexing binary
     *                      properties.
     * @return the lucene Document.
     * @throws RepositoryException if an error occurs while reading property
     *                             values from the <code>ItemStateProvider</code>.
     */
    public static Document createDocument(NodeState node,
                                          ItemStateManager stateProvider,
                                          NamespaceMappings mappings,
                                          List textFilters)
            throws RepositoryException {
        AdvNodeIndexer indexer = new AdvNodeIndexer(node, stateProvider, mappings, textFilters);
        return indexer.createDoc();
    }
    //private boolean addedToMaster;


    /**
     *
     * @param node
     * @param stateProvider
     * @param mappings
     * @param textFilters
     */
    protected AdvNodeIndexer(NodeState node, ItemStateManager stateProvider,
            NamespaceMappings mappings, List textFilters) {
        super(node, stateProvider, mappings, textFilters);
    }

    /**
     * Recursively searches for the parent node of the given node state. If a valid
     * parent is not found. <code>null</code> is returned.
     *
     * @param node
     * @param level
     * @return
     * @throws NoSuchItemStateException
     * @throws ItemStateException
     */
    private NodeState getDocMasterNode(NodeState node, int level)
            throws ItemStateException {

        final NodeState parent = getParentNodeState(node);
        if (parent != null) {
            // check if parent is the valid one
            if (level == 0) {
                // good to go
                log.info("master is: " + parent.getNodeTypeName());

                return parent;
            } else {
                return getDocMasterNode(parent, level - 1);
            }
        }
        //NodeState.ChildNodeEntry child = parent.getChildNodeEntry(node.getNodeId());
        //log.info("found master for: " + child);

        // there is no valid parent
        return null;
    }

    /**
     * Gets the parent node state of the given node state.
     * Returns <code>null</code> if there's none.
     *
     * @param node the current node state, must not be null
     * @return
     * @throws NoSuchItemStateException
     * @throws ItemStateException
     */
    private NodeState getParentNodeState(NodeState node)
            throws ItemStateException {
        assert node != null;
        if (node.getParentId() != null) {
            final NodeState parent = (NodeState) stateProvider
                    .getItemState(node.getParentId());

            return parent;
        }
        return null;
    }

    NodeState masterNode;

    static Set<String> masters = new HashSet<String>();

    /**
     * Creates a Lucene Document.
     *
     * @return the lucene Document with the index layout.
     * @throws RepositoryException if an error occurs while reading property
     *                             values from the <code>ItemStateProvider</code>.
     */
    protected synchronized Document createDoc() throws RepositoryException {
        Document doc = new Document();

        if (masters.contains(node.getNodeId().getUUID().toString())) {
            return null;
        }

        //addedToMaster = false;
        masterNode = null;
        // special fields
        // UUID
        //doc.add(new Field(FieldNames.UUID, node.getNodeId().getUUID().toString(), true, true, false));
        try {
            // parent UUID
            if (node.getParentId() == null) {
                // root node
                doc.add(new Field(FieldNames.PARENT, "", true, true, false));
                doc.add(new Field(FieldNames.LABEL, "", false, true, false));
            } else {
                doc.add(new Field(FieldNames.PARENT, node.getParentId().toString(), true, true, false));
                NodeState parent = (NodeState) stateProvider.getItemState(node.getParentId());
                NodeState.ChildNodeEntry child = parent.getChildNodeEntry(node.getNodeId());
                String name = child.getName().toJCRName(mappings);
                doc.add(new Field(FieldNames.LABEL, name, false, true, false));

            }
        } catch (NoSuchItemStateException e) {
            throwRepositoryException(e);
        } catch (ItemStateException e) {
            throwRepositoryException(e);
        } catch (NoPrefixDeclaredException e) {
            // will never happen, because this.mappings will dynamically add
            // unknown uri<->prefix mappings
        }

        // -- orig jackrabbit - start : duplicate in method __addNodeProps
        Set props = node.getPropertyNames();
        for (Object prop : props) {
            QName propName = (QName) prop;
            PropertyId id = new PropertyId(node.getNodeId(), propName);
            try {
                PropertyState propState = (PropertyState) stateProvider.getItemState(id);
                InternalValue[] values = propState.getValues();
                for (InternalValue value : values) {
                    addValue(doc, value, propState.getName());
                }
                if (values.length > 1) {
                    // real multi-valued
                    addMVPName(doc, propState.getName());
                }
            } catch (NoSuchItemStateException e) {
                throwRepositoryException(e);
            } catch (ItemStateException e) {
                throwRepositoryException(e);
            }
        }
        // -- orig jackrabbit - end

        //if (!addedToMaster)
        //    doc.add(new Field(FieldNames.UUID, node.getNodeId().getUUID().toString(), true, true, false));
        if (masterNode != null) {
            doc.add(new Field(FieldNames.UUID, masterNode.getNodeId().getUUID().toString(), true, true, false));

            try {
                if (masterNode.getParentId() != null) {
                    doc.add(new Field(FieldNames.PARENT, masterNode.getParentId().toString(), true, true, false));
                    NodeState parent = (NodeState) stateProvider.getItemState(masterNode.getParentId());
                    NodeState.ChildNodeEntry child = parent.getChildNodeEntry(node.getNodeId());
                    String name = child.getName().toJCRName(mappings);
                    doc.add(new Field(FieldNames.LABEL, name, false, true, false));
                }
            } catch (Exception e) {
                e.printStackTrace();
            }

            __addNodeProps(masterNode, doc);

            masters.add(masterNode.getNodeId().getUUID().toString());
        } else {
            doc.add(new Field(FieldNames.UUID, node.getNodeId().getUUID().toString(), true, true, false));
        }

        return doc;
    }

    /**
     * Actually indexes all other properties of a given node into the current doc.
     * The given node is a master Node (of a NX Document) that will be so linked
     * to the full text indexer.
     *
     * @param altnode
     * @param doc
     * @throws RepositoryException
     */
    private void __addNodeProps(NodeState altnode, Document doc) throws RepositoryException {
        Set props = altnode.getPropertyNames();
        for (Object prop : props) {
            QName propName = (QName) prop;
            PropertyId id = new PropertyId(altnode.getNodeId(), propName);
            try {
                PropertyState propState = (PropertyState) stateProvider.getItemState(id);
                InternalValue[] values = propState.getValues();
                for (InternalValue value : values) {
                    addValue(doc, value, propState.getName());
                }
                if (values.length > 1) {
                    // real multi-valued
                    addMVPName(doc, propState.getName());
                }
            } catch (NoSuchItemStateException e) {
                throwRepositoryException(e);
            } catch (ItemStateException e) {
                throwRepositoryException(e);
            }
        }
    }

    /**
     * Adds the index also for the master node of the JCRDocument which
     * is parent-parent of the current content node.
     */
    @Override
    protected void addBinaryValue(Document doc, String fieldName, Object internalValue) {
        final String logPrefix = "<addBinaryValue> ";
        log.info(logPrefix + "...");

        // TODO Auto-generated method stub
        //super.addBinaryValue(doc, fieldName, internalValue);

        //----------
        // 'check' if node is of type nt:resource
        try {
            String jcrData = mappings.getPrefix(QName.NS_JCR_URI) + ":data";
            if (!jcrData.equals(fieldName)) {
                // don't know how to index
                log.debug(logPrefix + "don't know how to index");
                return;
            }
            if (node.hasPropertyName(QName.JCR_MIMETYPE)) {
                PropertyState dataProp = (PropertyState) stateProvider.getItemState(
                        new PropertyId(node.getNodeId(), QName.JCR_DATA));
                PropertyState mimeTypeProp =
                        (PropertyState) stateProvider.getItemState(
                                new PropertyId(node.getNodeId(), QName.JCR_MIMETYPE));

                // jcr:encoding is not mandatory
                String encoding = null;
                if (node.hasPropertyName(QName.JCR_ENCODING)) {
                    PropertyState encodingProp =
                            (PropertyState) stateProvider.getItemState(
                                    new PropertyId(node.getNodeId(), QName.JCR_ENCODING));
                    encoding = encodingProp.getValues()[0].internalValue().toString();
                }

                String mimeType = mimeTypeProp.getValues()[0].internalValue().toString();
                Map fields = Collections.EMPTY_MAP;
                for (Object textFilter : textFilters) {
                    TextFilter filter = (TextFilter) textFilter;
                    // use the first filter that can handle the mimeType
                    if (filter.canFilter(mimeType)) {
                        fields = filter.doFilter(dataProp, encoding);
                        break;
                    }
                }

                // got the wordkeys... fields
                // ============================================================
                // NX specific code - START
                //
                // add also the parent-parent-... of type ecmnt:document to
                // the index target for this content
                log.info("Doc MASTER for: " + node.getNodeTypeName());
                NodeState masterNode = getDocMasterNode(node, 0);
                if (masterNode != null) {
                    //doc.removeField(arg0)
                    System.out.println("A  : " + node.getNodeId().getUUID());
                    System.out.println("B  : " + masterNode.getNodeId().getUUID());

                    //doc.add(new Field(FieldNames.UUID, masterNode.getNodeId().getUUID().toString(), true, true, false));

                    //__addNodeProps(masterNode, doc);
                    //doc.add(new Field(FieldNames.PARENT, "", true, true, false));
                    //doc.add(new Field(FieldNames.LABEL, "KK", false, true, false));

                    if (this.masterNode == null) {
                        this.masterNode = masterNode;
                    } else {
                        assert false : "master node not null";
                    }
                    //addedToMaster = true;
                }
                //
                // NX specific code - END
                // ============================================================

                for (Object o : fields.keySet()) {
                    String field = (String) o;
                    Reader r = (Reader) fields.get(field);
                    doc.add(Field.Text(field, r));
                }
            }
        } catch (ItemStateException e) {
            log.warn("Exception while indexing binary property: " + e.toString());
            log.debug("Dump: ", e);
        } catch (RepositoryException e) {
            log.warn("Exception while indexing binary property: " + e.toString());
            log.debug("Dump: ", e);
        }

    }
}
